@@ -4,23 +4,22 @@ import sys
 from collections import OrderedDict
 from collections.abc import Sequence
 from pathlib import Path
-from typing import Any, Optional, Union
+from typing import Any
 
 import matplotlib.pyplot as plt
+import numpy as np
 import scanpy as sc
 import spatialdata as sd
 from anndata import AnnData
 from dask.dataframe.core import DataFrame as DaskDataFrame
 from geopandas import GeoDataFrame
 from matplotlib.axes import Axes
-from matplotlib.colors import Colormap, Normalize
+from matplotlib.colors import Colormap, ListedColormap, Normalize
 from matplotlib.figure import Figure
 from multiscale_spatial_image.multiscale_spatial_image import MultiscaleSpatialImage
 from pandas.api.types import is_categorical_dtype
 from spatial_image import SpatialImage
-from spatialdata import transform
 from spatialdata._logging import logger as logg
-from spatialdata.transformations import get_transformation
 
 from spatialdata_plot._accessor import register_spatial_data_accessor
 from spatialdata_plot.pl.render import (
@@ -34,18 +33,18 @@ from spatialdata_plot.pl.render import (
     _render_shapes,
 )
 from spatialdata_plot.pl.utils import (
+    CmapParams,
     LegendParams,
-    Palette_t,
     _FontSize,
     _FontWeight,
     _get_cs_contents,
     _get_extent,
     _maybe_set_colors,
-    _multiscale_to_image,
+    _mpl_ax_contains_elements,
     _prepare_cmap_norm,
     _prepare_params_plot,
+    _robust_transform,
     _set_outline,
-    _translate_image,
     save_fig,
 )
 from spatialdata_plot.pp.utils import _verify_plotting_tree
@@ -84,11 +83,11 @@ class PlotAccessor:
 
     def _copy(
         self,
-        images: Union[None, dict[str, Union[SpatialImage, MultiscaleSpatialImage]]] = None,
-        labels: Union[None, dict[str, Union[SpatialImage, MultiscaleSpatialImage]]] = None,
-        points: Union[None, dict[str, DaskDataFrame]] = None,
-        shapes: Union[None, dict[str, GeoDataFrame]] = None,
-        table: Union[None, AnnData] = None,
+        images: None | dict[str, SpatialImage | MultiscaleSpatialImage] = None,
+        labels: None | dict[str, SpatialImage | MultiscaleSpatialImage] = None,
+        points: None | dict[str, DaskDataFrame] = None,
+        shapes: None | dict[str, GeoDataFrame] = None,
+        table: None | AnnData = None,
     ) -> sd.SpatialData:
         """Copy the current `SpatialData` object, optionally modifying some of its attributes.
 
@@ -141,20 +140,21 @@ class PlotAccessor:
 
     def render_shapes(
         self,
-        element: str | None = None,
+        elements: str | list[str] | None = None,
         color: str | None = None,
         groups: str | Sequence[str] | None = None,
         size: float = 1.0,
         outline: bool = False,
-        outline_width: tuple[float, float] = (0.3, 0.05),
-        outline_color: tuple[str, str] = ("black", "white"),
+        outline_width: float = 1.5,
+        outline_color: str | list[float] = "#000000ff",
         alt_var: str | None = None,
         layer: str | None = None,
-        palette: Palette_t = None,
+        palette: ListedColormap | str | None = None,
         cmap: Colormap | str | None = None,
-        norm: Optional[Normalize] = None,
+        norm: None | Normalize = None,
         na_color: str | tuple[float, ...] | None = "lightgrey",
-        alpha: float = 1.0,
+        outline_alpha: float = 1.0,
+        fill_alpha: float = 1.0,
         **kwargs: Any,
     ) -> sd.SpatialData:
         """
@@ -162,8 +162,8 @@ class PlotAccessor:
 
         Parameters
         ----------
-        element
-            The name of the shapes element to render. If `None`, the first
+        elements
+            The name of the shapes element(s) to render. If `None`, all
             shapes element in the `SpatialData` object will be used.
         color
             Key for annotations in :attr:`anndata.AnnData.obs` or variables/genes.
@@ -195,6 +195,11 @@ class PlotAccessor:
         kwargs
             Additional arguments to be passed to cmap and norm.
 
+        Notes
+        -----
+            Empty geometries will be removed at the time of plotting.
+            An ``outline_width`` of 0.0 leads to no border being plotted.
+
         Returns
         -------
         None
@@ -210,7 +215,7 @@ class PlotAccessor:
         )
         outline_params = _set_outline(size, outline, outline_width, outline_color)
         sdata.plotting_tree[f"{n_steps+1}_render_shapes"] = ShapesRenderParams(
-            element=element,
+            elements=elements,
             color=color,
             groups=groups,
             outline_params=outline_params,
@@ -218,20 +223,22 @@ class PlotAccessor:
             layer=layer,
             cmap_params=cmap_params,
             palette=palette,
-            alpha=alpha,
+            outline_alpha=outline_alpha,
+            fill_alpha=fill_alpha,
+            transfunc=kwargs.get("transfunc", None),
         )
 
         return sdata
 
     def render_points(
         self,
-        element: str | None = None,
+        elements: str | list[str] | None = None,
         color: str | None = None,
         groups: str | Sequence[str] | None = None,
         size: float = 1.0,
-        palette: Palette_t = None,
+        palette: ListedColormap | str | None = None,
         cmap: Colormap | str | None = None,
-        norm: Optional[Normalize] = None,
+        norm: None | Normalize = None,
         na_color: str | tuple[float, ...] | None = (0.0, 0.0, 0.0, 0.0),
         alpha: float = 1.0,
         **kwargs: Any,
@@ -241,8 +248,8 @@ class PlotAccessor:
 
         Parameters
         ----------
-        element
-            The name of the points element to render. If `None`, the first
+        elements
+            The name of the points element(s) to render. If `None`, all
             shapes element in the `SpatialData` object will be used.
         color
             Key for annotations in :attr:`anndata.AnnData.obs` or variables/genes.
@@ -278,25 +285,28 @@ class PlotAccessor:
             **kwargs,
         )
         sdata.plotting_tree[f"{n_steps+1}_render_points"] = PointsRenderParams(
-            element=element,
+            elements=elements,
             color=color,
             groups=groups,
             cmap_params=cmap_params,
             palette=palette,
             alpha=alpha,
+            transfunc=kwargs.get("transfunc", None),
+            size=size,
         )
 
         return sdata
 
     def render_images(
         self,
-        element: str | None = None,
+        elements: str | list[str] | None = None,
         channel: list[str] | list[int] | int | str | None = None,
-        cmap: Colormap | str | None = None,
-        norm: Optional[Normalize] = None,
+        cmap: list[Colormap] | list[str] | Colormap | str | None = None,
+        norm: None | Normalize = None,
         na_color: str | tuple[float, ...] | None = (0.0, 0.0, 0.0, 0.0),
-        palette: Palette_t = None,
+        palette: ListedColormap | str | None = None,
         alpha: float = 1.0,
+        quantiles_for_norm: tuple[float | None, float | None] = (3.0, 99.8),  # defaults from CSBDeep
         **kwargs: Any,
     ) -> sd.SpatialData:
         """
@@ -304,9 +314,9 @@ class PlotAccessor:
 
         Parameters
         ----------
-        element
-            The name of the image element to render. If `None`, the first
-            shapes element in the `SpatialData` object will be used.
+        elements
+            The name of the image element(s) to render. If `None`, all
+            shapes elements in the `SpatialData` object will be used.
         channel
             To select which channel to plot (all by default).
         cmap
@@ -317,6 +327,8 @@ class PlotAccessor:
             Color to be used for NAs values, if present.
         alpha
             Alpha value for the shapes.
+        quantiles_for_norm
+            Tuple of (pmin, pmax) which will be used for quantile normalization.
         kwargs
             Additional arguments to be passed to cmap and norm.
 
@@ -328,34 +340,52 @@ class PlotAccessor:
         sdata = _verify_plotting_tree(sdata)
         n_steps = len(sdata.plotting_tree.keys())
 
-        cmap_params = _prepare_cmap_norm(
-            cmap=cmap,
-            norm=norm,
-            na_color=na_color,  # type: ignore[arg-type]
-            **kwargs,
-        )
+        if channel is None and cmap is None:
+            cmap = "brg"
+
+        cmap_params: list[CmapParams] | CmapParams
+        if isinstance(cmap, list):
+            cmap_params = [
+                _prepare_cmap_norm(
+                    cmap=c,
+                    norm=norm,
+                    na_color=na_color,  # type: ignore[arg-type]
+                    **kwargs,
+                )
+                for c in cmap
+            ]
+
+        else:
+            cmap_params = _prepare_cmap_norm(
+                cmap=cmap,
+                norm=norm,
+                na_color=na_color,  # type: ignore[arg-type]
+                **kwargs,
+            )
+
         sdata.plotting_tree[f"{n_steps+1}_render_images"] = ImageRenderParams(
-            element=element,
+            elements=elements,
             channel=channel,
             cmap_params=cmap_params,
             palette=palette,
             alpha=alpha,
+            quantiles_for_norm=quantiles_for_norm,
         )
 
         return sdata
 
     def render_labels(
         self,
-        element: str | None = None,
+        elements: str | list[str] | None = None,
         color: str | None = None,
         groups: str | Sequence[str] | None = None,
         contour_px: int = 3,
         outline: bool = False,
         alt_var: str | None = None,
         layer: str | None = None,
-        palette: Palette_t = None,
+        palette: ListedColormap | str | None = None,
         cmap: Colormap | str | None = None,
-        norm: Optional[Normalize] = None,
+        norm: None | Normalize = None,
         na_color: str | tuple[float, ...] | None = (0.0, 0.0, 0.0, 0.0),
         outline_alpha: float = 1.0,
         fill_alpha: float = 0.3,
@@ -366,9 +396,9 @@ class PlotAccessor:
 
         Parameters
         ----------
-        element
-            The name of the labels element to render. If `None`, the first
-            labels element in the `SpatialData` object will be used.
+        elements
+            The name of the labels element(s) to render. If `None`, all
+            labels elements in the `SpatialData` object will be used.
         color
             Key for annotations in :attr:`anndata.AnnData.obs` or variables/genes.
         groups
@@ -417,7 +447,7 @@ class PlotAccessor:
             **kwargs,
         )
         sdata.plotting_tree[f"{n_steps+1}_render_labels"] = LabelsRenderParams(
-            element=element,
+            elements=elements,
             color=color,
             groups=groups,
             contour_px=contour_px,
@@ -428,6 +458,7 @@ class PlotAccessor:
             palette=palette,
             outline_alpha=outline_alpha,
             fill_alpha=fill_alpha,
+            transfunc=kwargs.get("transfunc", None),
         )
 
         return sdata
@@ -448,9 +479,12 @@ class PlotAccessor:
         figsize: tuple[float, float] | None = None,
         dpi: int | None = None,
         fig: Figure | None = None,
+        title: None | str | Sequence[str] = None,
+        share_extent: bool = True,
+        pad_extent: int = 0,
         ax: Axes | Sequence[Axes] | None = None,
         return_ax: bool = False,
-        save: Optional[Union[str, Path]] = None,
+        save: None | str | Path = None,
     ) -> sd.SpatialData:
         """
         Plot the images in the SpatialData object.
@@ -505,60 +539,67 @@ class PlotAccessor:
                 render_cmds[cmd] = params
 
         if len(render_cmds.keys()) == 0:
-            raise TypeError("Please specify what to plot using the 'render_*' functions before calling 'imshow().")
+            raise TypeError("Please specify what to plot using the 'render_*' functions before calling 'imshow()'.")
 
-        # Simplicstic solution: If the images are multiscale, just use the first
-        sdata = _multiscale_to_image(sdata)
+        if title is not None:
+            if isinstance(title, str):
+                title = [title]
 
-        img_transformations: dict[str, dict[str, sd.transformations.transformations.BaseTransformation]] = {}
-        # transform all elements
-        for cmd, _ in render_cmds.items():
-            if cmd == "render_images" or cmd == "render_channels":
-                for key in sdata.images:
-                    img_transformations[key] = {}
-                    all_transformations = get_transformation(sdata.images[key], get_all=True)
+            if not all(isinstance(t, str) for t in title):
+                raise TypeError("All titles must be strings.")
 
-                    for cs, transformation in all_transformations.items():
-                        img_transformations[key][cs] = transformation
+        # get original axis extent for later comparison
+        x_min_orig, x_max_orig = (np.inf, -np.inf)
+        y_min_orig, y_max_orig = (np.inf, -np.inf)
 
-                        if isinstance(transformation, sd.transformations.transformations.Translation):
-                            sdata.images[key] = _translate_image(image=sdata.images[key], translation=transformation)
-
-                        elif isinstance(transformation, sd.transformations.transformations.Sequence):
-                            # we have more than one transformation, let's find the translation(s)
-                            for t in list(transformation.transformations):
-                                if isinstance(t, sd.transformations.transformations.Translation):
-                                    sdata.images[key] = _translate_image(image=sdata.images[key], translation=t)
-
-                                else:
-                                    sdata.images[key] = transform(sdata.images[key], t)
-
-            elif cmd == "render_shapes":
-                for key in sdata.shapes:
-                    shape_transformation = get_transformation(sdata.shapes[key], get_all=True)
-                    shape_transformation = list(shape_transformation.values())[0]
-                    sdata.shapes[key] = transform(sdata.shapes[key], shape_transformation)
-
-            elif cmd == "render_labels":
-                for key in sdata.labels:
-                    label_transformation = get_transformation(sdata.labels[key], get_all=True)
-                    label_transformation = list(label_transformation.values())[0]
-                    sdata.labels[key] = transform(sdata.labels[key], label_transformation)
-
-        extent = _get_extent(
-            sdata=sdata,
-            images="render_images" in render_cmds,
-            labels="render_labels" in render_cmds,
-            points="render_points" in render_cmds,
-            shapes="render_shapes" in render_cmds,
-            img_transformations=img_transformations if len(img_transformations) > 0 else None,
-        )
+        if isinstance(ax, Axes) and _mpl_ax_contains_elements(ax):
+            x_min_orig, x_max_orig = ax.get_xlim()
+            y_max_orig, y_min_orig = ax.get_ylim()  # (0, 0) is top-left
 
         # handle coordinate system
         coordinate_systems = sdata.coordinate_systems if coordinate_systems is None else coordinate_systems
         if isinstance(coordinate_systems, str):
             coordinate_systems = [coordinate_systems]
 
+        for cs in coordinate_systems:
+            if cs not in sdata.coordinate_systems:
+                raise ValueError(f"Unknown coordinate system '{cs}', valid choices are: {sdata.coordinate_systems}")
+
+        # Check if user specified only certain elements to be plotted
+        cs_contents = _get_cs_contents(sdata)
+        elements_to_be_rendered = []
+        for cmd, params in render_cmds.items():
+            if cmd == "render_images" and cs_contents.query(f"cs == '{cs}'")["has_images"][0]:  # noqa: SIM114
+                if params.elements is not None:
+                    elements_to_be_rendered += (
+                        [params.elements] if isinstance(params.elements, str) else params.elements
+                    )
+            elif cmd == "render_shapes" and cs_contents.query(f"cs == '{cs}'")["has_shapes"][0]:  # noqa: SIM114
+                if params.elements is not None:
+                    elements_to_be_rendered += (
+                        [params.elements] if isinstance(params.elements, str) else params.elements
+                    )
+            elif cmd == "render_points" and cs_contents.query(f"cs == '{cs}'")["has_points"][0]:  # noqa: SIM114
+                if params.elements is not None:
+                    elements_to_be_rendered += (
+                        [params.elements] if isinstance(params.elements, str) else params.elements
+                    )
+            elif cmd == "render_labels" and cs_contents.query(f"cs == '{cs}'")["has_labels"][0]:  # noqa: SIM102
+                if params.elements is not None:
+                    elements_to_be_rendered += (
+                        [params.elements] if isinstance(params.elements, str) else params.elements
+                    )
+
+        extent = _get_extent(
+            sdata=sdata,
+            has_images="render_images" in render_cmds,
+            has_labels="render_labels" in render_cmds,
+            has_points="render_points" in render_cmds,
+            has_shapes="render_shapes" in render_cmds,
+            elements=elements_to_be_rendered,
+            coordinate_systems=coordinate_systems,
+        )
+
         # Use extent to filter out coordinate system without the relevant elements
         valid_cs = []
         for cs in coordinate_systems:
@@ -568,14 +609,6 @@ class PlotAccessor:
                 logg.info(f"Dropping coordinate system '{cs}' since it doesn't have relevant elements.")
         coordinate_systems = valid_cs
 
-        # check that coordinate system and elements to be rendered match
-        for cmd, params in render_cmds.items():
-            if params.element is not None and len([params.element]) != len(coordinate_systems):
-                raise ValueError(
-                    f"Number of coordinate systems ({len(coordinate_systems)}) does not match number of elements "
-                    f"({len(params.element)}) in command {cmd}."
-                )
-
         # set up canvas
         fig_params, scalebar_params = _prepare_params_plot(
             num_panels=len(coordinate_systems),
@@ -598,9 +631,29 @@ class PlotAccessor:
         )
 
         # go through tree
-        cs_contents = _get_cs_contents(sdata)
         for i, cs in enumerate(coordinate_systems):
+            sdata = self._copy()
+            # properly transform all elements to the current coordinate system
+            members = cs_contents.query(f"cs == '{cs}'")
+
+            if members["has_images"].values[0]:
+                for key in sdata.images:
+                    sdata.images[key] = _robust_transform(sdata.images[key], cs)
+
+            if members["has_labels"].values[0]:
+                for key in sdata.labels:
+                    sdata.labels[key] = _robust_transform(sdata.labels[key], cs)
+
+            if members["has_points"].values[0]:
+                for key in sdata.points:
+                    sdata.points[key] = _robust_transform(sdata.points[key], cs)
+
+            if members["has_shapes"].values[0]:
+                for key in sdata.shapes:
+                    sdata.shapes[key] = _robust_transform(sdata.shapes[key], cs)
+
             ax = fig_params.ax if fig_params.axs is None else fig_params.axs[i]
+
             for cmd, params in render_cmds.items():
                 if cmd == "render_images" and cs_contents.query(f"cs == '{cs}'")["has_images"][0]:
                     _render_images(
@@ -611,6 +664,7 @@ class PlotAccessor:
                         fig_params=fig_params,
                         scalebar_params=scalebar_params,
                         legend_params=legend_params,
+                        # extent=extent[cs],
                     )
                 elif cmd == "render_shapes" and cs_contents.query(f"cs == '{cs}'")["has_shapes"][0]:
                     if sdata.table is not None and isinstance(params.color, str):
@@ -644,11 +698,7 @@ class PlotAccessor:
                     )
 
                 elif cmd == "render_labels" and cs_contents.query(f"cs == '{cs}'")["has_labels"][0]:
-                    if (
-                        sdata.table is not None
-                        # and isinstance(params["instance_key"], str)
-                        and isinstance(params.color, str)
-                    ):
+                    if sdata.table is not None and isinstance(params.color, str):
                         colors = sc.get.obs_df(sdata.table, params.color)
                         if is_categorical_dtype(colors):
                             _maybe_set_colors(
@@ -667,17 +717,33 @@ class PlotAccessor:
                         legend_params=legend_params,
                     )
 
-                ax.set_title(cs)
-                if any(
-                    [
-                        cs_contents.query(f"cs == '{cs}'")["has_images"][0],
-                        cs_contents.query(f"cs == '{cs}'")["has_labels"][0],
-                        cs_contents.query(f"cs == '{cs}'")["has_points"][0],
-                        cs_contents.query(f"cs == '{cs}'")["has_shapes"][0],
-                    ]
-                ):
-                    ax.set_xlim(extent[cs][0], extent[cs][1])
-                    ax.set_ylim(extent[cs][3], extent[cs][2])  # (0, 0) is top-left
+                if title is not None:
+                    if len(title) == 1:
+                        t = title[0]
+                    else:
+                        try:
+                            t = title[i]
+                        except IndexError as e:
+                            raise IndexError("The number of titles must match the number of coordinate systems.") from e
+                else:
+                    t = cs
+                ax.set_title(t)
+
+            if any(
+                [
+                    cs_contents.query(f"cs == '{cs}'")["has_images"][0],
+                    cs_contents.query(f"cs == '{cs}'")["has_labels"][0],
+                    cs_contents.query(f"cs == '{cs}'")["has_points"][0],
+                    cs_contents.query(f"cs == '{cs}'")["has_shapes"][0],
+                ]
+            ):
+                # If the axis already has limits, only expand them but not overwrite
+                x_min = min(x_min_orig, extent[cs][0]) - pad_extent
+                x_max = max(x_max_orig, extent[cs][1]) + pad_extent
+                y_min = min(y_min_orig, extent[cs][2]) - pad_extent
+                y_max = max(y_max_orig, extent[cs][3]) + pad_extent
+                ax.set_xlim(x_min, x_max)
+                ax.set_ylim(y_max, y_min)  # (0, 0) is top-left
 
         if fig_params.fig is not None and save is not None:
             save_fig(fig_params.fig, path=save)
