@@ -10,6 +10,7 @@ import matplotlib
 import numpy as np
 import pandas as pd
 import scanpy as sc
+import spatial_image
 import spatialdata as sd
 from anndata import AnnData
 from geopandas import GeoDataFrame
@@ -19,6 +20,10 @@ from matplotlib.colors import ColorConverter, ListedColormap, Normalize
 from matplotlib.patches import Circle, Polygon
 from pandas.api.types import is_categorical_dtype
 from scanpy._settings import settings as sc_settings
+from spatialdata.models import (
+    Image2DModel,
+    Labels2DModel,
+)
 
 from spatialdata_plot._logging import logger
 from spatialdata_plot.pl.utils import (
@@ -116,6 +121,8 @@ def _render_shapes(
     ) -> PatchCollection:
         patches = []
         for shape in shapes:
+            # remove empty points/polygons
+            shape = shape[shape["geometry"].apply(lambda geom: not geom.is_empty)]
             # We assume that all elements in one collection are of the same type
             if shape["geometry"].iloc[0].geom_type == "Polygon":
                 patches += [Polygon(p.exterior.coords, closed=True) for p in shape["geometry"]]
@@ -149,7 +156,7 @@ def _render_shapes(
             patches,
             snap=False,
             # zorder=4,
-            lw=1.5,
+            lw=render_params.outline_params.linewidth,
             facecolor=fill_c,
             edgecolor=outline_c,
             **kwargs,
@@ -342,8 +349,11 @@ def _render_images(
         elements = list(sdata_filt.images.keys())
 
     images = [sdata.images[e] for e in elements]
+    for img, img_key in zip(images, elements):
+        if not isinstance(img, spatial_image.SpatialImage):
+            img = Image2DModel.parse(img["scale0"].ds.to_array().squeeze(axis=0))
+            logger.warning(f"Multi-scale images not yet supported, using scale0 of multi-scale image '{img_key}'.")
 
-    for img in images:
         if render_params.channel is None:
             channels = img.coords["c"].values
         else:
@@ -512,9 +522,13 @@ def _render_labels(
     if elements is None:
         elements = list(sdata_filt.labels.keys())
 
-    labels = [sdata.labels[e].values for e in elements]
+    labels = [sdata.labels[e] for e in elements]
+
+    for label, label_key in zip(labels, elements):
+        if not isinstance(label, spatial_image.SpatialImage):
+            label = Labels2DModel.parse(label["scale0"].ds.to_array().squeeze(axis=0))
+            logger.warning(f"Multi-scale labels not yet supported, using scale0 of multi-scale label '{label_key}'.")
 
-    for label, labels_key in zip(labels, elements):
         if sdata.table is None:
             instance_id = np.unique(label)
             table = AnnData(None, obs=pd.DataFrame(index=np.arange(len(instance_id))))
@@ -522,7 +536,7 @@ def _render_labels(
             instance_key = _get_instance_key(sdata)
             region_key = _get_region_key(sdata)
 
-            table = sdata.table[sdata.table.obs[region_key].isin([labels_key])]
+            table = sdata.table[sdata.table.obs[region_key].isin([label_key])]
 
             # get isntance id based on subsetted table
             instance_id = table.obs[instance_key].values
@@ -542,7 +556,7 @@ def _render_labels(
         if (render_params.fill_alpha != render_params.outline_alpha) and render_params.contour_px is not None:
             # First get the labels infill and plot them
             labels_infill = _map_color_seg(
-                seg=label,
+                seg=label.values,
                 cell_id=instance_id,
                 color_vector=color_vector,
                 color_source_vector=color_source_vector,
@@ -565,7 +579,7 @@ def _render_labels(
 
             # Then overlay the contour
             labels_contour = _map_color_seg(
-                seg=label,
+                seg=label.values,
                 cell_id=instance_id,
                 color_vector=color_vector,
                 color_source_vector=color_source_vector,
